
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory

import h5py
import numpy as np
import pytest

from cal_tools.calcat_interface2 import JUNGFRAUConditions
from cal_tools.constants import write_ccv


def test_write_ccv():
    # Create a temporary CCV file for testing
    with TemporaryDirectory() as temp_dir:
        ccv_path = Path(temp_dir) / "test.ccv"
        conditions = JUNGFRAUConditions(
            sensor_bias_voltage=180,
            memory_cells=1,
            integration_time=10,
            exposure_timeout=25,
            gain_setting=0,
            gain_mode=0,
        )
        # Create test data
        data = np.random.rand(10, 10)
        dims = ["x", "y"]
        pdu_name = "Jungfrau_M275"
        pdu_uuid = 140
        detector_type = "test-detector"
        calibration = "Offset10Hz"
        created_at = datetime.strptime(
            "2022-10-08T00:07:06.000+02:00", "%Y-%m-%dT%H:%M:%S.%f%z")
        proposal = 1234
        runs = [1, 2, 3]

        write_ccv(
            ccv_path, pdu_name, pdu_uuid, detector_type,
            calibration, conditions, created_at, proposal, runs,
            data, dims)
        # Check the CCV file was created with the expected data
        with h5py.File(ccv_path, "r") as f:
            assert f.attrs["version"] == 0
            pdu_group = f[pdu_name]
            assert pdu_group.attrs["uuid"] == pdu_uuid
            assert pdu_group.attrs["detector_type"] == detector_type
            ccv_group = pdu_group[calibration]["0"]
            assert ccv_group.attrs["begin_at"] == created_at.isoformat()
            assert ccv_group.attrs["proposal"] == proposal
            np.testing.assert_array_equal(ccv_group.attrs["runs"], runs)
            np.testing.assert_array_equal(ccv_group["data"], data)
            np.testing.assert_array_equal(
                ccv_group["data"].attrs["dims"], dims)


def test_write_ccv_invalid_dims():
    # Test that ValueError is raised if data.ndim != len(dims)
    with pytest.raises(ValueError):
        write_ccv(
            None, None, None, None, None, None, None, None, None,
            np.array([1, 2, 3]), ["x", "y"])
