from __future__ import annotations

import argparse
import datetime as dt
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Generator, Optional

import zmq
from config import webservice as config
from httpx import Client, Response
from rich import print
from rich.progress import (
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)
from rich.prompt import Confirm

parser = argparse.ArgumentParser(
    description="Manually submit calibration jobs.",
    epilog="""To run in the background use `nohup PYTHONUNBUFFERED=1 python
    manual_launch.py ... &` followed by `disown`.""",
)

parser.add_argument(
    "slices",
    type=str,
    nargs="+",
    help="""slices (or single numbers) of runs to process, inclusive range, starting at
    1 (e.g. 1:3 parsed to {1, 2, 3}, 10 parsed to {10}, :10 parsed to {1, 2, ...,
    10})""",
)
parser.add_argument(
    "--proposal",
    type=int,
    help="proposal number",
    required=True,
)
parser.add_argument(
    "--delay",
    default=30,
    type=int,
    help="delay in seconds between submissions",
    required=False,
)
parser.add_argument(
    "--noconfirm",
    action="store_true",
    help="skip confirmation",
)
parser.add_argument(
    "--really",
    action="store_true",
    help="actually submit jobs instead of just printing them",
)


BEARER = {
    "access_token": "",
    "expires_at": dt.datetime.now(),
}


def pre_checks():
    # Fail fast if we don't have the required configs set
    required_keys = ["token-url", "user-id", "user-secret", "user-email"]
    for k in required_keys:
        if config["metadata-client"][k] is None:
            print(
                f"Missing key [bold red]`{k}`[/bold red] in metadata client configuration"
            )
            print("[bold red]Aborted[/bold red]")
            exit(1)


def get_bearer_token() -> str:
    if BEARER["access_token"] and BEARER["expires_at"] > dt.datetime.now():
        return BEARER["access_token"]

    with Client() as client:
        data = {
            "grant_type": "client_credentials",
            "client_id": config["metadata-client"]["user-id"],
            "client_secret": config["metadata-client"]["user-secret"],
        }

        # With a real user the scope must be set to public
        if not str(config["metadata-client"]["user-email"]).endswith("example.com"):
            data["scope"] = "public"

        response = client.post(f"{config['metadata-client']['token-url']}", data=data)

    data = response.json()

    if any(k not in data for k in ["access_token", "expires_in"]):
        print(
            "Response from MyMdC missing required fields, check webservice `user-id`"
            f"and `user-secret`. Response: {data=}",
        )
        raise ValueError("Invalid response from MyMdC")

    expires_in = dt.timedelta(seconds=data["expires_in"])

    BEARER["access_token"] = data["access_token"]
    BEARER["expires_at"] = dt.datetime.now() + expires_in

    return BEARER["access_token"]


@contextmanager
def get_client() -> Generator[Client, None, None]:
    bearer_token = get_bearer_token()

    with Client() as client:
        headers = {
            "accept": "application/json; version=1",
            "X-User-Email": config["metadata-client"]["user-email"],
            "Authorization": f"Bearer {bearer_token}",
        }

        client.headers.update(headers)

        yield client


def _get_runs_by_proposal(number: int, client: Client, page: int = 1) -> Response:
    return client.get(
        f"{config['metadata-client']['base-api-url']}/runs/runs_by_proposal",
        params={"proposal_number": number, "page": page},
        timeout=10,
    )


def get_runs_by_proposal_all(number: int) -> list[dict]:
    with get_client() as client:
        res = _get_runs_by_proposal(number, client, 1)
        if res.status_code != 200:
            raise ValueError(res.url, res.text)
        runs = res.json()
        for page in range(2, int(res.headers.get("x-total-pages", 1)) + 1):
            _ = _get_runs_by_proposal(number, client, page)
            runs.extend(_.json())

    return runs


def main(
    proposal_no: int,
    slices: list[slice],
    delay: int,
    noconfirm: Optional[bool] = False,
    really: Optional[bool] = False,
):
    with Progress(transient=True) as progress:
        task_submission = progress.add_task(
            "[yellow]Querying FS for proposal information", total=None
        )
        exp = Path("/gpfs/exfel/exp")
        proposal_paths = list(exp.glob(f"*/*/p{proposal_no:06d}"))
        if len(proposal_paths) != 1:
            raise ValueError(f"Proposal {proposal_no} not found")

        proposal_path = proposal_paths[0]
        instrument = proposal_path.parts[4]
        cycle = proposal_path.parts[5]

        progress.update(task_submission, description="[yellow]Querying MyMdC for runs")

        all_runs = get_runs_by_proposal_all(proposal_no)

    run_no_id_map = {run["run_number"]: run["id"] for run in all_runs}
    max_run_no = max(run_no_id_map.keys()) + 1
    requested_ranges = [range(*s.indices(max_run_no)) for s in slices]
    requested_run_nos = {run_no for r in requested_ranges for run_no in r}

    requests = dict(
        sorted(
            {
                run_no: run_no_id_map[run_no]
                for run_no in requested_run_nos
                if run_no in run_no_id_map
            }.items()
        )
    )

    if missing_run_ids := set(requested_run_nos) - set(run_no_id_map.keys()):
        print(
            f"[bold red]Missing run IDs for run number(s) {missing_run_ids}[/bold red]"
        )

    if not really:
        print("[yellow]`--really` flag missing, not submitting jobs")

    if not noconfirm and not Confirm.ask(
        f"Submit [red bold]{len(requests)}[/red bold] jobs for proposal "
        f"[bold]{proposal_no}[/bold]?",
        default=False,
    ):
        print("[bold red]Aborted[/bold red]")
        exit(1)

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
    ) as progress:
        description = f"[green]Submitted request for p{proposal_no:05d}/{{run_str}} "
        task_submission = progress.add_task(
            f"{description}r---[------]", total=len(requests)
        )
        con = zmq.Context()
        socket = con.socket(zmq.REQ)
        socket.connect("tcp://max-exfl-cal001:5555")

        if not really:
            #  Fake socket for testing, just logs what would have been sent via ZMQ
            socket = lambda: None
            socket.send = lambda x: progress.console.log(
                f"mock `zmq.REQ` socket send: {x}"
            )
            socket.recv = lambda: "mock `zmq.REQ` socket response"

        last_run_no = list(requests.keys())[-1]

        for run_no, run_id in requests.items():
            args = (
                "correct",
                str(run_id),
                "_",
                str(instrument),
                str(cycle),
                f"{proposal_no:06d}",
                str(run_no),
                "-",
            )
            msg = f"""['{"','".join(args)}']""".encode()
            progress.console.log(args)
            socket.send(msg)

            progress.update(
                task_submission,
                advance=1,
                description=description.format(
                    run_str=f"[bold yellow]r{run_no:03d}[{run_id:06d}]"
                ),
            )

            res = socket.recv()
            progress.console.log(res)

            if run_no != last_run_no:
                progress.console.log(f"sleeping for {delay}s")
                time.sleep(delay)
            else:
                progress.update(task_submission, description="[green]Done")


if __name__ == "__main__":
    args = vars(parser.parse_args())

    slices = []
    for s in args["slices"]:
        slice_split = tuple(map(lambda x: int(x) if x else None, s.split(":")))
        sep = None
        if len(slice_split) == 1:
            start, stop = slice_split[0], slice_split[0]
        elif len(slice_split) == 2:
            start, stop = slice_split
        else:
            start, stop, sep = slice_split

        # Python slice indices are 0-based, but we want to be 1-based
        if start is None or start == 0:
            start = 1

        if stop:
            stop = stop + 1 if stop != -1 else stop

        slices.append(slice(start, stop, sep))

    pre_checks()

    main(
        args["proposal"],
        slices,
        args["delay"],
        args["noconfirm"],
        args["really"],
    )
